
export iers_obliquity

"""
    iers_obliquity(m::IERSModel, tt_c::Number)

Compute the mean obliquity of the ecliptic at epoch, in radians, at time `tt_c` expressed in 
`TT` Julian centuries since `J2000`, according to the IERS convention `m`.

!!! note 
    The mean obliquity for the IERS 2003 conventions already accounts the adjustment to the 
    IAU 1976 precession model for the IAU 2000 precession-rate of the equator in obliquity .

### References 
- IERS Technical Note No. [21](https://www.iers.org/IERS/EN/Publications/TechnicalNotes/tn21.html)
- IERS Technical Note No. [32](https://www.iers.org/IERS/EN/Publications/TechnicalNotes/tn32.html)
- IERS Technical Note No. [36](https://www.iers.org/IERS/EN/Publications/TechnicalNotes/tn36.html) 
"""
function iers_obliquity(::IERS1996, tt_c::Number)
    return arcsec2rad(@evalpoly(tt_c, 84381.448, -46.8150, -0.00059, 0.001813))
end

function iers_obliquity(::IERS2003, tt_c::Number)
    δωₐ = -0.02524 # IAU 2003 Precession-rate adjustment
    return arcsec2rad(@evalpoly(tt_c, 84381.448, -46.8150 + δωₐ, -0.00059, 0.001813))
end

function iers_obliquity(::IERS2010, tt_c::Number)
    return arcsec2rad(        
        @evalpoly(
            tt_c, 84381.406, -46.836769, -0.0001831, 0.00200340, -0.000000576, -0.0000000434
        )
    )
end